from pyPot import PyPot 
from vec3 import norm, Vec3 
from atomSel import AtomSel
import numpy as np
import atomAction 
from math import pi, sqrt
from dihedral import Dihedral
import protocol


from vec3 import dot, cross, norm, unitVec
from math import atan2, pi

class TwistPot( PyPot ) :
    
    def __init__(self, name, anglelist): 
        PyPot.__init__(self ,name) #first call base class constructor 
        self.resetPotName('TwistPot')
        self.setThreshold(2)  # default threshold in degrees
        
        self.restraints_=[] 
        for r in   anglelist:
            obs=r[4]  
            self.restraints_.append(Restraint(r[0],r[1],r[2],r[3], obs,"",self) )        
        return
        
    def restraints(self):
        return self.restraints_
    
    def numRestraints(self):
        return len(self.restraints())
        
    def violations(self):
        viols=0
        for r in self.restraints():
            if r.violated():
                viols += 1
            else:
                viols += 0 
        return viols
    
    def rms(self):
        """RMSD in degrees
        """
        #s.calcEnergy()
        ret=0.
        for r in self.restraints():
            diff=(r.value()*(180./pi)) - r.obs*(180./pi) #in degrees
            ret += diff**2
            pass
        from math import sqrt
        if self.numRestraints()>0:
            ret = sqrt(ret/self.numRestraints())
            pass
        return ret
    
    
    def setScale(self, scale):
        self.scale=scale
        return scale
    
    def calcEnergy(self):
        ene=0.
        forcec=self.setScale(self.scale)
        
        for r in self.restraints():
            ene += r.Energy(forcec)
        return ene
    
    def clear(self):
        del self.restraints_[:]

    def calcEnergyAndDerivList(self, derivs):
        #calcEnergyAndDerivList(self, derivs) creates a derivList object named derivs
        #    indexed by simulation, do derivList.get() to get the gradient list per atom 
        
        #calcEnergyAndDerivs(self, derivs) creates the gradient list named derivs indexed by atom
        #both functions retrun the energy!
        forcec=self.setScale(self.scale)
        
        import xplor
        sim = self.restraints()[0].atom1.simulation()
        
        from derivList import DerivList
        dlist = DerivList()
        dlist.init() 
 
        for r in self.restraints():            
            #scale=(2.*k*(self.value()-self.obs))
            r.deriv(dlist, forcec)

        #dlist[sim].scale(s.scale())
        derivs[sim] += dlist[sim]
        
        return self.calcEnergy()  
        
pass

class Restraint:
    def __init__(self,atom1,atom2,atom3,atom4, obs,comment,pot):
        """
        obs is input in degrees
        """
        self.pot=pot
        from atomSel import AtomSel
        
        self.atom1 = AtomSel(atom1)[0]
        self.atom2 = AtomSel(atom2)[0]
        self.atom3 = AtomSel(atom3)[0]
        self.atom4 = AtomSel(atom4)[0]
       
        self.obs=obs*(pi/180.0)
        self.comment=comment
        return
       
    def printRemarks(self): #fix
        remark=  "------------------------------------------------------------\n"
        remark +="""TWISTpot               Value\nTWISTpot            %.3f """ %(self.value()*(180./pi))
        return remark

    def string(self):
        s='(%s), (%s), (%s), (%s)' %(self.atom1.string(), self.atom2.string(), self.atom3.string(), self.atom4.string())
        return s
    
    def calc(self):
        L=(self.atom1.pos()+ self.atom2.pos())/2.
        M=(self.atom2.pos()+ self.atom3.pos())/2.
        N=(self.atom3.pos()+ self.atom4.pos())/2.
        
        P=(L+ M)/2.
        Q=(M+ N)/2.
    
        v1 = self.atom2.pos()-P
        v2 = Q-P
        v3 = -v2
        v4 = self.atom3.pos()-Q
        
        v12 = cross(v1,v2)
        v34 = cross(v3,v4)

        phi=atan2(dot(unitVec(v2) , cross(v12, v34)) , dot( v12, v34))
        return phi

    def value(self):
       return self.calc()#/(pi/180.0)
    
    def violated(self):
        diff=(self.value()*(180./pi)) - self.obs*(180./pi) #in degrees
        if abs(diff) > self.pot.threshold():
            return True
        else:
            return False
   
    def Energy(self, forcec):
        a=self.value()
        b=self.obs
        if a>b:
           return forcec*((a-b)**1) 
        if a<b:
           return forcec*((b-a)**1) 

    def deriv(self, dlist, forcec):
        L=(self.atom1.pos()+ self.atom2.pos())/2.
        M=(self.atom2.pos()+ self.atom3.pos())/2.
        N=(self.atom3.pos()+ self.atom4.pos())/2.
        
        P=(L+ M)/2.
        Q=(M+ N)/2.
    
        v1 = self.atom2.pos()-P
        v2 = Q-P
        v3 = -v2
        v4 = self.atom3.pos()-Q
        
        v1c2 = cross(v1,v2)
        v2c4 = cross(v2,v4)
        v3c4 = cross(v3,v4)
        v4c1 = cross(v4,v1)
        v1d2 = dot(v1,v2)
        v1d4 = dot(v1,v4)
        v2d4 = dot(v2,v4)
        v3d4 = dot(v3,v4)

        u1 = unitVec(v1c2)
        u2 = unitVec(v3c4)
        u1d2 = dot(u1,u2)
        u1c2 = cross(u1,u2)

        n2 = norm(v2)

        N = 1./(n2*norm(v1c2)*norm(v3c4))

        phi=atan2(dot(unitVec(v2) , u1c2) , u1d2)
        #print 'phi:' ,phi * 180/pi


        dPhid0 = -u1d2*n2**2*v2c4
        dPhid0 -= n2*dot(unitVec(v2),u1c2) * n2**2*v4
        dPhid0 += n2*dot(unitVec(v2),u1c2) * v2d4*v2
        
        dPhid3 = -u1d2*n2**2*v1c2 - \
                 n2*dot(unitVec(v2),u1c2) * (n2**2*v1-v1d2*v2)

        dPhid1 = -dPhid0 + u1d2*(n2**2*v4c1 + 2*dot(v1c2,v4)*v2) +\
                 dot(unitVec(v2), u1c2)*(3*v1d4*n2*v2 - n2*v2d4*v1 - \
                                         n2*v1d2*v4-v1d2*v2d4*unitVec(v2))

        dPhid2 = -dPhid3 - u1d2*(n2**2*v4c1 + 2*dot(v1c2,v4)*v2) -\
                 dot(unitVec(v2), u1c2)*(3*v1d4*n2*v2 - n2*v2d4*v1 - \
                                         n2*v1d2*v4-v1d2*v2d4*unitVec(v2))
        
        a=self.value()
        b=self.obs
        k=forcec
        
        if a>b:
            k=k
        if a<b:
            k=-k
           
        scale=k                                 
        
        dlist[self.atom2] +=  scale* dPhid0*N/4.
        dlist[self.atom1] +=  scale* dPhid1*N/4.
        dlist[self.atom4] +=  scale* dPhid2*N/4.
        dlist[self.atom3] +=  scale* dPhid3*N/4.
        
        return    
    pass


def analyze(potList):
    "perform analysis of Pot terms and return nicely formatted summary"

    ret = ""

    from simulationTools import getPotTerms
    potList = getPotTerms(potList,'TwistPot')
    if not potList: return ret

    instanceNames = map(lambda x: x.instanceName(),potList)
    instanceNames.sort()

    for name in instanceNames:
        term = filter(lambda x: x.instanceName()==name,potList)[0]
        for r in term.restraints():
            print r.value()

        pass
    
    return ret
#
import simulationTools
simulationTools.registerTerm(analyze,
                             "TwistPot Analysis","TwistPot",
r"""
For each term report the calculated, target value and difference in angle.
""")
        


